from collections import defaultdict
import json


class SchemaJSON:
    """Schema for JSON objects
    """

    def __new__(cls):
        obj = super(SchemaJSON, cls).__new__(cls)
        obj._field_dict = {}
        return obj

    def __setattr__(self, name, value):

        if type(value) is DeserField or type(value) is DeserObjectField:

            if value.name is None:
                value.name = value.name_conv(name)

            if name in self._field_dict:
                self._field_dict[name].deserialize = value
            else:
                self._field_dict[name] = CompositeField(deserialize=value)

            self.__dict__[name] = None

        elif type(value) is SerField or type(value) is SerObjectField:

            # If name is not specified use the name of the variable
            if value.name is None:
                value.name = value.name_conv(name)

            if name in self._field_dict:
                self._field_dict[name].serialize = value
            else:
                self._field_dict[name] = CompositeField(serialize=value)

            self.__dict__[name] = None

        else:
            object.__setattr__(self, name, value)


class BaseJSON:
    """Base class for JSON objects
    """

    def set_schema_json(self, schema):
        """Sets the schema for this object
        """
        self._field_dict = schema._field_dict

    def to_json(self, filename=None):
        """Serializes the object into a JSON file.
        """
        json_dict = self.to_dict()

        if filename is None:
            return json.dumps(json_dict)

        with open(filename, "w") as f:
            f.write(json.dumps(json_dict))

    def to_dict(self):
        """Converts this object into a python dict"""
        data_dict = {}
        for var_name, field in self._field_dict.items():
            serialize = field.serialize

            if self.__dict__.get(var_name, serialize.default) is None:
                if serialize.optional:
                    continue
                else:
                    raise Exception('var "{}" is None'.format(var_name))

            sub_data_dict = data_dict
            # iterate through parent keys until target key
            for key in serialize.parent_keys:
                if key in sub_data_dict:
                    if type(sub_data_dict[key]) is dict:
                        sub_data_dict = sub_data_dict[key]
                    else:
                        raise Exception(
                            ('parent key "{}" ' "is populated").format(key)
                        )
                else:
                    sub_data_dict[key] = {}
                    sub_data_dict = sub_data_dict[key]

            if type(serialize) is SerField:
                kind = serialize.kind
                if serialize.repeated:
                    sub_data_dict[serialize.name] = []
                    for value in self.__dict__.get(var_name):
                        sub_data_dict[serialize.name].append(kind(value))
                else:
                    json_value = kind(
                        self.__dict__.get(var_name, serialize.default)
                    )
                    sub_data_dict[serialize.name] = json_value
            elif type(serialize) is SerObjectField:
                if serialize.repeated:
                    sub_data_dict[serialize.name] = []
                    for obj in self.__dict__[var_name]:
                        sub_data_dict[serialize.name].append(obj.to_dict())
                else:
                    obj = self.__dict__[var_name]
                    sub_data_dict[serialize.name] = obj.to_dict()

        return data_dict

    def from_json(self, filename=None, raw_json=None):
        """Loads in values to this object from a JSON file or string"""
        if filename is not None:
            with open(filename, "r") as f:
                raw_json = f.read()
        elif raw_json is None:
            raise Exception("Specify filename or raw JSON")

        data_dict = json.loads(raw_json)
        self.from_dict(data_dict)

    def from_dict(self, data_dict):
        """Loads in values to this object from a python dict"""
        for var_name, field in self._field_dict.items():
            deserialize = field.deserialize
            sub_data_dict = data_dict

            for key in deserialize.parent_keys:
                if key in sub_data_dict:
                    sub_data_dict = sub_data_dict[key]
                elif deserialize.optional:
                    continue
                else:
                    raise Exception(
                        (
                            'parent_key "{}" does not exists for ' "field {}"
                        ).format(key, deserialize.name)
                    )

            if deserialize.name not in sub_data_dict:
                if deserialize.optional:
                    continue
                raise Exception(
                    "{} field not found in the json".format(deserialize.name)
                )

            if type(deserialize) is DeserField:
                if deserialize.repeated:
                    self.__dict__[var_name] = []
                    for value in sub_data_dict[deserialize.name]:
                        self.__dict__[var_name].append(deserialize.kind(value))
                else:
                    self.__dict__[var_name] = deserialize.kind(
                        sub_data_dict[deserialize.name]
                    )
            elif type(deserialize) is DeserObjectField:
                if deserialize.repeated:
                    self.__dict__[var_name] = []
                    for value in sub_data_dict[deserialize.name]:
                        obj = deserialize.kind()
                        obj.from_dict(value)
                        self.__dict__[var_name].append(obj)
                else:
                    self.__dict__[var_name] = deserialize.kind()
                    self.__dict__[var_name].from_dict(
                        sub_data_dict[deserialize.name]
                    )


class CompositeField:
    """Class for holding both serialize and deserialize fields"""

    def __init__(self, serialize=None, deserialize=None):
        self.serialize = serialize
        self.deserialize = deserialize


class Field:
    """Super class for fields"""

    def __str__(self):
        return "(name: {0}, kind: {1})".format(self.name, self.kind)

    __repr__ = __str__


class DeserField(Field):
    """DeserField
    name:
        name of the field that should be deserialized from, default is the
        name of the variable
    kind:
        the type of the variable should be deserialized to, default is what
        the deserialized variable is
    optional:
        specifies if the value is optional, if value is not found as a field
        the variable is left as None
    parent_keys
        The parent keys of this field.
    """

    def __init__(
        self,
        name=None,
        name_conv=lambda x: x,
        kind=lambda x: x,
        optional=False,
        repeated=False,
        parent_keys=[],
    ):
        self.name = name
        self.name_conv = name_conv
        self.kind = kind
        self.optional = optional
        self.repeated = repeated
        self.parent_keys = parent_keys

        if not callable(kind):
            raise Exception("Kind needs to be callable")


class SerField(Field):
    """SerField
    name:
        name of the field that should serialize to, defaults to the variable
        name
    kind:
        the type that the field should be serialized to, defaults to the type
        that the variable is
    optional:
        specifies if the value is optional, if value is None then it will not
        get serialized
    parent_keys
        The parent keys of this field.
    default
        The default value of this field. Field must be optional to have
        default value.
    """

    def __init__(
        self,
        name=None,
        name_conv=lambda x: x,
        kind=lambda x: x,
        optional=False,
        repeated=False,
        parent_keys=[],
        default=None,
    ):
        self.name = name
        self.name_conv = name_conv
        self.kind = kind
        self.optional = optional
        self.repeated = repeated
        self.parent_keys = parent_keys
        self.default = default

        if not callable(kind):
            raise Exception("Kind needs to be callable")


class ObjectField:
    """Super class for object fields"""

    def __str__(self):
        return "(name: {0}, repeated: {1})".format(self.name, self.repeated)

    __repr__ = __str__


class SerObjectField(ObjectField):
    """SerObjectField
    name:
        name of the field that should serialize to, defaults to the variable
        name
    kind:
        the type that the field should be serialized to, defaults to the type
        that the variable is
    optional:
        specifies if the value is optional, if value is None then it will not
        get serialized
    parent_keys
        The parent keys of this field.

    """

    def __init__(
        self,
        name=None,
        name_conv=lambda x: x,
        optional=False,
        repeated=False,
        parent_keys=[],
        default=None,
    ):
        self.name = name
        self.name_conv = name_conv
        self.optional = optional
        self.repeated = repeated
        self.parent_keys = parent_keys
        self.default = default

        if not optional and default is not None:
            raise Exception("Default value populated while optional")

        if not type(parent_keys) is list:
            raise Exception("parent keys has to be of type list")


class DeserObjectField(ObjectField):
    """DeserObjectField

    name
        The name of the field that should contain this object.
    optional
        Specifies if this field is optional.
    repeated
        Specifies if this field contains a list.
    kind
        Specifies which class should be instantiated for this field.
    parent_keys
        The parent keys of this field.
    """

    def __init__(
        self,
        name=None,
        name_conv=lambda x: x,
        optional=False,
        repeated=False,
        kind=None,
        parent_keys=[],
    ):

        self.name = name
        self.name_conv = name_conv
        self.optional = optional
        self.repeated = repeated
        self.kind = kind
        self.parent_keys = parent_keys



from .json import (
    BaseJSON,
    SchemaJSON,
    SerField,
    DeserField,
    DeserObjectField,
    SerObjectField,
)
from .configparser import (
    ConfigSectionBase,
    ConfigBase,
    SerConfigOption,
    DeserConfigOption,
    SerConfigSection,
    DeserConfigSection,
)



import configparser


class ConfigBaseParent(object):
    # TODO: add checks for if filename is valid
    def to_config(self, filename=None):
        config = configparser.ConfigParser()

        for field_name, field in self._config_key_dict.items():
            serialize = field.serialize
            if serialize is None:
                continue

            value = self.__dict__[field_name]
            if value is None and serialize.optional:
                continue

            value = serialize.kind(value)

            if not config.has_section(serialize.section):
                config.add_section(serialize.section)
            config.set(serialize.section, serialize.name, value)

        if filename is not None:
            with open(filename, "w") as f:
                config.write(f)
        else:
            return config

    def from_config(self, filename):
        config = configparser.ConfigParser()
        config.read(filename)
        for field_name, field in self._config_key_dict.items():
            deserialize = field.deserialize
            if deserialize is None:
                continue

            if type(field) is CompositeConfigOption:
                if config.has_option(deserialize.section, deserialize.name):
                    self.__dict__[field_name] = field.deserialize.kind(
                        config.get(deserialize.section, deserialize.name)
                    )
                elif (
                    not config.has_section(deserialize.section)
                    and not deserialize.optional
                ):
                    raise Exception(
                        '"{}" section not in config'.format(deserialize.section)
                    )
                elif not deserialize.optional:
                    raise Exception(
                        '"{}" option not in config'.format(deserialize.name)
                    )
            elif type(field) is CompositeConfigSection:
                self.__dict__[field_name].from_config(filename)


class ConfigSectionBase(ConfigBaseParent):
    """ConfigSectionBase
    """

    def __new__(cls):
        obj = super(ConfigSectionBase, cls).__new__(cls)
        obj._config_key_dict = {}
        return obj

    def __setattr__(self, name, value):
        if type(value) is DeserConfigOption:
            if value.name is None:
                value.name = name

            if name in self._config_key_dict:
                self._config_key_dict[name].deserialize = value
            else:
                self._config_key_dict[name] = CompositeConfigOption(
                    deserialize=value
                )
            self.__dict__[name] = None

        elif type(value) is SerConfigOption:
            if value.name is None:
                value.name = name

            if name in self._config_key_dict:
                self._config_key_dict[name].serialize = value
            else:
                self._config_key_dict[name] = CompositeConfigOption(
                    serialize=value
                )
            self.__dict__[name] = None
        else:
            object.__setattr__(self, name, value)

    def __set_serialize_section__(self, section):
        for field in self._config_key_dict.values():
            if field.serialize is not None:
                field.serialize.section = section

    def __set_deserialize_section__(self, section):
        for field in self._config_key_dict.values():
            if field.deserialize is not None:
                field.deserialize.section = section


class ConfigBase(ConfigBaseParent):
    def __new__(cls):
        obj = super(ConfigBase, cls).__new__(cls)
        obj._config_key_dict = {}
        return obj

    def __setattr__(self, name, value):
        if type(value) is DeserConfigOption:
            if value.name is None:
                value.name = name
            if name in self._config_key_dict:
                self._config_key_dict[name].deserialize = value
            else:
                self._config_key_dict[name] = CompositeConfigOption(
                    deserialize=value
                )
            self.__dict__[name] = None
        elif type(value) is DeserConfigSection:
            obj = value.kind()
            obj.__set_deserialize_section__(value.section)
            if name in self._config_key_dict:
                self._config_key_dict[name].deserialize = value
            else:
                self._config_key_dict[name] = CompositeConfigSection(
                    deserialize=value
                )
            self.__dict__[name] = obj

        elif type(value) is SerConfigOption:
            if value.name is None:
                value.name = name

            if name in self._config_key_dict:
                self._config_key_dict[name].serialize = value
            else:
                self._config_key_dict[name] = CompositeConfigOption(
                    serialize=value
                )
            self.__dict__[name] = None

        elif type(value) is SerConfigSection:
            obj = value.kind()
            obj.__set_serialize_section__(value.section)
            if name in self._config_key_dict:
                self._config_key_dict[name].serialize = value
            else:
                self._config_key_dict[name] = CompositeConfigSection(
                    serialize=value
                )
            self.__dict__[name] = obj
        else:
            object.__setattr__(self, name, value)


class CompositeConfigOption:
    def __init__(self, serialize=None, deserialize=None, is_section=False):
        self.serialize = serialize
        self.deserialize = deserialize

    def __str__(self):
        return "serialize: {} deserialize: {}".format(
            self.serialize, self.deserialize
        )

    __repr__ = __str__


class ConfigOption:
    def __str__(self):
        return "(section {0} name: {1}, kind: {2}, optional: {3})".format(
            self.section, self.name, self.kind, self.optional
        )

    __repr__ = __str__


class SerConfigOption(ConfigOption):
    def __init__(self, name=None, section=None, kind=str, optional=False):
        self.name = name
        self.section = section
        self.kind = kind
        self.optional = optional


class DeserConfigOption(ConfigOption):
    def __init__(
        self, name=None, section=None, kind=lambda x: x, optional=False
    ):
        self.name = name
        self.section = section
        self.kind = kind
        self.optional = optional


class CompositeConfigSection:
    def __init__(self, serialize=None, deserialize=None, is_section=False):
        self.serialize = serialize
        self.deserialize = deserialize

    def __str__(self):
        return "serialize: {} deserialize: {}".format(
            self.serialize, self.deserialize
        )

    __repr__ = __str__


class ConfigSection:
    def __str__(self):
        return "(section {0} optional: {1})".format(self.section, self.optional)

    __repr__ = __str__


class SerConfigSection(ConfigSection):
    def __init__(self, kind, section, optional=False):
        self.kind = kind
        self.section = section
        self.optional = optional


class DeserConfigSection(ConfigSection):
    def __init__(self, kind, section, optional=False):
        self.kind = kind
        self.section = section
        self.optional = optional



import sys
from codecs import open

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

MAJOR = 0
MINOR = 1
MICRO = 4
VERSION = "%d.%d.%d" % (MAJOR, MINOR, MICRO)


class PyTest(TestCommand):
    """Handle test execution from setup."""

    user_options = [("pytest-args=", "a", "Arguments to pass into pytest")]

    def initialize_options(self):
        """Initialize the PyTest options."""
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def finalize_options(self):
        """Finalize the PyTest options."""
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        """Run the PyTest testing suite."""
        try:
            import pytest
        except ImportError:
            raise ImportError(
                "Running tests requires additional dependencies."
                "\nPlease run (pip install pytest[test])"
            )

        errno = pytest.main(self.pytest_args.split(" "))
        sys.exit(errno)


cmdclass = {"test": PyTest}  # Define custom commands.


if "build_docs" in sys.argv:
    try:
        from sphinx.setup_command import BuildDoc
    except ImportError:
        raise ImportError(
            "Running the documenation builds has additional"
            " dependencies. Please run (pip install pyser[docs])"
        )
    cmdclass["build_docs"] = BuildDoc


# Define the requirements for specific execution needs.
requires = []

test_reqs = [
    "pytest-cov>=2.5.1",
    "pytest>=3.0.0",
    "coveralls>=1.1,<2.0",
    "rstcheck>=3.3.1",
]

doc_reqs = [
    "sphinx_rtd_theme>=0.1.10b0S",
    "Sphinx>=1.5.2",
]

extra_reqs = {"doc": doc_reqs, "test": test_reqs}

with open("README.rst", "r", "utf-8") as fh:
    long_description = fh.read()


setup(
    name="pyser",
    version=VERSION,
    author="Jonne Kaunisto",
    author_email="jonneka@gmail.com",
    description="Python Serializer and Deserializer",
    long_description=long_description,
    url="https://github.com/jonnekaunisto/pyser",
    license="MIT License",
    keywords="serialize, deserialize",
    packages=find_packages(exclude=["docs", "test"]),
    cmdclass=cmdclass,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Utilities"
    ],
    install_requires=requires,
    tests_require=test_reqs,
    extras_require=extra_reqs,
)
